#include "externalFunc.h"
#include "externalAggFunc.h"
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <oci.h>
#include <odci.h>

typedef struct GetNLQImpl
{
  OCIRaw* key;
}GetNLQImpl;

typedef struct GetNLQImpl_ind
{
  OCIInd _atomic;
  OCIInd key;
}GetNLQImpl_ind;

typedef struct Handles_t
{
  OCIExtProcContext* extProcCtx;
  OCIEnv* envhp;
  OCISvcCtx* svchp;
  OCIError* errhp;
  OCISession* usrhp;
}Handles_t;

typedef struct StoredCtx
{
  int d;
  double n;
  double L[MAX_D];
  double Q[MAX_D][MAX_D];
}StoredCtx;

static int checkerr(Handles_t* handles, sword status)
{
  text errbuf[512];     // error message buffer 
  sb4 errcode;          // OCI error code 
  
  switch (status)
  {
  case OCI_SUCCESS:
  case OCI_SUCCESS_WITH_INFO:
    return 0;
  case OCI_ERROR:
    OCIErrorGet ((dvoid*) handles->errhp, (ub4) 1, (text *) NULL, &errcode,
                 errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR);
    sprintf((char*)errbuf, "OCI ERROR code %d",errcode);
    break;
  default:
    sprintf((char*)errbuf, "Warning - error status %d",status);
    break;
  }

  OCIExtProcRaiseExcpWithMsg(handles->extProcCtx, 29400, errbuf,
    strlen((char*)errbuf));

  return -1;
}

static int printerr(Handles_t* handles, sword status, text *errbuf)
{
  sb4 errcode;          // OCI error code 

  switch (status)
  {
  case OCI_SUCCESS:
  case OCI_SUCCESS_WITH_INFO:
    return 0;
  case OCI_ERROR:
    OCIErrorGet ((dvoid*) handles->errhp, (ub4) 1, (text *) NULL, &errcode,
                 errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR);
    sprintf((char*)errbuf, "OCI ERROR code %d",errcode);
    break;
  default:
    sprintf((char*)errbuf, "Warning - error status %d",status);
    break;
  }

  return -1;
}

static int GetHandles(OCIExtProcContext* extProcCtx, Handles_t* handles)
{
  /* store the ext-proc context in the handles struct */
  handles->extProcCtx=extProcCtx;

  /* Get OCI handles */
  if (checkerr(handles, OCIExtProcGetEnv(extProcCtx, &handles->envhp,
                          &handles->svchp, &handles->errhp)))
     return -1;

  /* get the user handle */
  if (checkerr(handles, OCIAttrGet((dvoid*)handles->svchp,
                                   (ub4)OCI_HTYPE_SVCCTX, 
                                   (dvoid*)&handles->usrhp,
                                   (ub4*) 0, (ub4)OCI_ATTR_SESSION, 
                                   handles->errhp)))
     return -1;

  return 0;
}

static StoredCtx* GetStoredCtx(Handles_t* handles, GetNLQImpl* self, 
                               GetNLQImpl_ind* self_ind)
{
  StoredCtx *storedCtx;           /* Stored context pointer */
  ub1 *key;                       /* key to retrieve context */
  ub4 keylen;                     /* length of key */
  
  /* return NULL if the PL/SQL context is NULL */
  if (self_ind->_atomic == OCI_IND_NULL) return NULL;

  /* Get the key */
  key = OCIRawPtr(handles->envhp, self->key);
  keylen = OCIRawSize(handles->envhp, self->key);
  /* Retrieve stored context using the key */
  if (checkerr(handles, OCIContextGetValue((dvoid*) handles->usrhp, 
                                           handles->errhp,
                                           key, (ub1) keylen, 
                                           (dvoid**) &storedCtx)))
     return NULL;

  return storedCtx;
}

DLLIMPORT OCINumber* ODCIAggregateInitialize(OCIExtProcContext* context,
                                             GetNLQImpl* sctx,
                                             GetNLQImpl_ind* sctx_ind,
                                             short* ret_ind)
{
    int i,j;
	StoredCtx* storedCtx;                /* Stored context pointer */
	ub4 key;                             /* key to retrieve stored context */
	Handles_t handles;
	
	GetHandles(context,&handles);
	OCIMemoryAlloc((dvoid*) handles.usrhp, 
                   handles.errhp,
                   (dvoid**) &storedCtx,
                   OCI_DURATION_STATEMENT,
                   (ub4) sizeof(StoredCtx),
                   OCI_MEMORY_CLEARED);
    
    storedCtx->d = -1;               
    storedCtx->n=0;
    for(i=0;i<MAX_D;i++){
        storedCtx->L[i] = 0;
        for(j=0;j<MAX_D;j++)
            storedCtx->Q[i][j] = 0;
    }
    
	OCIContextGenerateKey((dvoid*) handles.usrhp, handles.errhp, &key);
	OCIContextSetValue((dvoid*)handles.usrhp, 
						handles.errhp,
						OCI_DURATION_STATEMENT,
                        (ub1*) &key,
						(ub1) sizeof(key),
                        (dvoid*) storedCtx);
	OCIRawAssignBytes(	handles.envhp, 
						handles.errhp, 
						(ub1*) &key, 
						(ub4) sizeof(key),
						&(sctx->key));
	sctx_ind->_atomic=OCI_IND_NOTNULL;
	sctx_ind->key=OCI_IND_NOTNULL;
	OCINumber * returned;
	int success = ODCI_SUCCESS;
	OCIMemoryAlloc((dvoid*) handles.usrhp, 
					handles.errhp,
                    (dvoid**) &returned,
                    OCI_DURATION_STATEMENT,
                    (ub4) sizeof(OCINumber),
                    OCI_MEMORY_CLEARED);
	OCINumberFromInt( handles.errhp,
                      &success, 
                      sizeof(int),
                      OCI_NUMBER_SIGNED,
                      returned );
                      
    *ret_ind = (short)OCI_IND_NOTNULL;
	return returned;
}

DLLIMPORT OCINumber* ODCIAggregateIterate(OCIExtProcContext *context,
                                          GetNLQImpl* self,
                                          GetNLQImpl_ind* self_ind,
                                          char* str, 
                                          short str_ind, 
                                          int str_len,
                                          short* ret_ind)
{
    double a[MAX_D];
    int sizeA;
    int i,j;
    
	StoredCtx* storedCtx;
	Handles_t  handles;
	OCINumber* returned;
	
	GetHandles(context,&handles);
	if(str_ind==OCI_IND_NOTNULL){
       storedCtx=GetStoredCtx(&handles,self,self_ind);
	   OCIMemoryAlloc((dvoid*) handles.usrhp, 
	   				  handles.errhp,
                      (dvoid**) &returned,
                      OCI_DURATION_STATEMENT,
                      (ub4) sizeof(OCINumber),
                      OCI_MEMORY_CLEARED);
    
       
       split(str,a,&sizeA);
       if(storedCtx->d == -1)
          storedCtx->d = sizeA;
       
       for(i=0;i<sizeA;i++){
           storedCtx->L[i]+= a[i];
           for(j=0;j<=i;j++)//Triangular Optimization
               storedCtx->Q[i][j]+=a[i]*a[j];
       }
       
       storedCtx->n++;
    }

	int success = ODCI_SUCCESS;
	OCINumberFromInt( handles.errhp,
                      &success, 
                      sizeof(int),
                      OCI_NUMBER_SIGNED,
                      returned );
                      
    *ret_ind = (short)OCI_IND_NOTNULL;
    
	return returned;
}

DLLIMPORT OCINumber* ODCIAggregateMerge(OCIExtProcContext* context,
                                        GetNLQImpl* self,
                                        GetNLQImpl_ind* self_ind,
                                        GetNLQImpl* ctx2,
                                        GetNLQImpl_ind* ctx2_ind,
                                        short* ret_ind)
{
    int d,i,j;
    StoredCtx* storedCtx1;
    StoredCtx* storedCtx2;
	Handles_t  handles;
	OCINumber* returned = NULL;
	
	GetHandles(context,&handles);
	storedCtx1=GetStoredCtx(&handles,self,self_ind);
    storedCtx2=GetStoredCtx(&handles,ctx2,ctx2_ind);
    
    d = storedCtx1->d;
    for(i=0;i<d;i++){
        storedCtx1->L[i] += storedCtx2->L[i];          
        for(j=0;j<=i;j++)//Triangular Optimization
            storedCtx1->Q[i][j] += storedCtx2->Q[i][j];
    }
    storedCtx1->n += storedCtx2->n;	
	OCIMemoryAlloc((dvoid*) handles.usrhp, 
	   				  handles.errhp,
                      (dvoid**) &returned,
                      OCI_DURATION_STATEMENT,
                      (ub4) sizeof(OCINumber),
                      OCI_MEMORY_CLEARED);
	
    int success = ODCI_SUCCESS;
	OCINumberFromInt( handles.errhp,
                      &success, 
                      sizeof(int),
                      OCI_NUMBER_SIGNED,
                      returned );
                      
	*ret_ind = (short)OCI_IND_NOTNULL;
	
    return returned;
    //DebugBreak();                                 
}                                            

DLLIMPORT OCINumber* ODCIAggregateTerminate(OCIExtProcContext* context,
                                            GetNLQImpl* self,
                                            GetNLQImpl_ind* self_ind,
                                            OCILobLocator** lob_ret,
                                            short* lob_ind,
                                            //char*  returnValue,
                                            //short* returnValue_ind,
                                            //int*   returnValue_len,
                                            //int*   retrunValue_maxlen,
                                            OCINumber*  flags,
                                            short* flags_ind,
                                            short* ret_ind
                                            )
{	
	StoredCtx* storedCtx;
	Handles_t  handles;
	OCINumber* returned = NULL;
	
    //char bufferL[(MAX_LOB_LEN/5)]="";
    //char bufferQ[(MAX_LOB_LEN/5)*4]="";
    char buffer [MAX_LOB_LEN]="";
    int i,j,d;

	GetHandles(context,&handles);
	storedCtx=GetStoredCtx(&handles,self,self_ind);
    
    //Funcin regresar CLOB -- Varchar solo maneja hasta 4000 bytes
    oraub8 buflen;
    oraub8 amt;
    
    /* Allocate a new lob descriptor for lob_out */
    OCIDescriptorAlloc((void *) handles.envhp,
                       (dvoid **) &lob_ret[0],
                       (ub4) OCI_DTYPE_LOB,       /* Type of Descriptor */
                       (size_t) 0,
                       (dvoid **) 0);
    
    OCILobCreateTemporary(handles.svchp, 
                          handles.errhp, 
                          lob_ret[0], (ub2)0,
                          (ub1)SQLCS_IMPLICIT, 
                          (ub1)OCI_TEMP_CLOB,
                          TRUE, 
                          OCI_DURATION_SESSION);
    /*                      
    char* str = "xxxx";
    int str_len = strlen(str);
    
    amt    = (oraub8)str_len; 
    buflen = (oraub8)str_len;  
                        
    OCILobWriteAppend2(handles.svchp,
                       handles.errhp, 
                       lob_ret[0], NULL,
                       &amt, (dvoid *)str, 
                       buflen,
                       OCI_ONE_PIECE, (void *)0,
                       (OCICallbackLobWrite2) 0,
                       0, SQLCS_IMPLICIT);
                       */
    //A. Pasar N a cadena (incluye separador_)
    sprintf(buffer,"%d_",(int)storedCtx->n);
    
    //1.- Se concatena N (buffer)                      
    amt    = (oraub8)strlen(buffer); 
    buflen = (oraub8)amt;  
                        
    OCILobWriteAppend2(handles.svchp,
                       handles.errhp, 
                       lob_ret[0], NULL,
                       &amt, (dvoid *)buffer, buflen,
                       OCI_ONE_PIECE, (void *)0,
                       (OCICallbackLobWrite2) 0,
                       0, SQLCS_IMPLICIT);
                       
    //B. Transformar L a cadena
    join(storedCtx->L,&storedCtx->d,buffer);
    
    //2. Se concatena L
    amt    = (oraub8)strlen(buffer); 
    buflen = (oraub8)amt;  
                        
    OCILobWriteAppend2(handles.svchp,
                       handles.errhp, 
                       lob_ret[0], NULL,
                       &amt, (dvoid *)buffer, buflen,
                       OCI_ONE_PIECE, (void *)0,
                       (OCICallbackLobWrite2) 0,
                       0, SQLCS_IMPLICIT); 
    
    //3. Se concatena separador
    amt    = (oraub8)strlen((char*)"_"); 
    buflen = (oraub8)amt;  
                        
    OCILobWriteAppend2(handles.svchp,
                       handles.errhp, 
                       lob_ret[0], NULL,
                       &amt, (dvoid *)((char*)"_"), buflen,
                       OCI_ONE_PIECE, (void *)0,
                       (OCICallbackLobWrite2) 0,
                       0, SQLCS_IMPLICIT);  
    
    //C. Transformar Q a cadena
    d = storedCtx->d; 
    for(i=0;i<d;i++){
       for(j=(i+1);j<=(d-1);j++)                   //Triangular Optimization
          storedCtx->Q[i][j] = storedCtx->Q[j][i];//Duplicate lower to upper triangular
       
       join(storedCtx->Q[i],&storedCtx->d,buffer);
       if(i!=storedCtx->d-1)
          strcat(buffer,(char*)"//");
       
       //strcat(bufferQ,buffer);
       //4. Se concatena Q
       amt    = (oraub8)strlen(buffer); 
       buflen = (oraub8)amt;  
                            
       OCILobWriteAppend2(handles.svchp,
                          handles.errhp, 
                          lob_ret[0], NULL,
                          &amt, (dvoid *)buffer, buflen,
                          OCI_ONE_PIECE, (void *)0,
                          (OCICallbackLobWrite2) 0,
                          0, SQLCS_IMPLICIT);
       
       
    }
    
    /* Set indicator of lob_ret */
    *lob_ind = (short)OCI_IND_NOTNULL;
    
    //strcat(returnValue,bufferL);
    //strcat(returnValue,(char*)"_");
    //strcat(returnValue,bufferQ);
    
    //*returnValue_len = strlen(returnValue);
	//*returnValue_ind = (short)OCI_IND_NOTNULL;
	
	
	OCIMemoryAlloc((dvoid*) handles.usrhp, 
                   handles.errhp,
                   (dvoid**) &returned,
                   OCI_DURATION_STATEMENT,
                   (ub4) sizeof(OCINumber),
                   OCI_MEMORY_CLEARED);
	
    int success = ODCI_SUCCESS;
	OCINumberFromInt( handles.errhp,
                      &success, 
                      sizeof(int),
                      OCI_NUMBER_SIGNED,
                      returned);
                      
	*ret_ind = (short)OCI_IND_NOTNULL;
	
    return returned;
}

                                            
DLLIMPORT  void getLob(OCIExtProcContext* context,
GetNLQImpl* self,
GetNLQImpl_ind* self_ind,
OCINumber*  flags,
short* flags_ind,
OCILobLocator** lob_ret,
short* lob_ind,
short* success_ind,
OCINumber* success

                                //char*  str,
                                //short  str_ind,
                                //int    str_len,
                                //short* ind_ret
                                )
{
    Handles_t      handles;
    //OCILobLocator* lob_ret;
    oraub8 buflen;
    oraub8 amt;
    
    char* str = "Holaa";
    int str_len = strlen(str);
    
    GetHandles(context,&handles);
    
    //
    OCIMemoryAlloc((dvoid*) handles.usrhp, 
                   handles.errhp,
                   (dvoid**) &success,
                   OCI_DURATION_STATEMENT,
                   (ub4) sizeof(OCINumber),
                   OCI_MEMORY_CLEARED);
	
    int success_val = ODCI_SUCCESS;
	OCINumberFromInt( handles.errhp,
                      &success_val, 
                      sizeof(int),
                      OCI_NUMBER_SIGNED,
                      success);
                      
	*success_ind = (short)OCI_IND_NOTNULL;
    //
    
    /* Allocate a new lob descriptor for lob_out */
    OCIDescriptorAlloc((void *) handles.envhp,
                       (dvoid **) &lob_ret[0],
                       (ub4) OCI_DTYPE_LOB,       /* Type of Descriptor */
                       (size_t) 0,
                       (dvoid **) 0);
    
    OCILobCreateTemporary(handles.svchp, 
                          handles.errhp, 
                          lob_ret[0], 0,
                          SQLCS_IMPLICIT, 
                          OCI_TEMP_CLOB,
                          TRUE, 
                          OCI_DURATION_SESSION);
    amt    = (oraub8)str_len; 
    buflen = (oraub8)str_len;  
                        
    OCILobWriteAppend2(handles.svchp,
                       handles.errhp, 
                       lob_ret[0], NULL,
                       &amt, (dvoid *)str, buflen,
                       OCI_ONE_PIECE, (void *)0,
                       (OCICallbackLobWrite2) 0,
                       0, SQLCS_IMPLICIT);
            
    /* Set indicator of lob_ret */
    *lob_ind = (short)OCI_IND_NOTNULL;

    /* Return lob_ret locator back to PL/SQL */
    return;
    //return(lob_ret);
}
